# copyright ############################### #
# This file is part of the Xwakes Package.  #
# Copyright (c) CERN, 2024.                 #
# ######################################### #

REL_TOL = 1e-8
ABS_TOL = 1e-8
PRECISION_FACTOR = 100.

MIN_FREQ = 1e3
MAX_FREQ = 1e10
FREQ_P_FACTOR = 0.2

MIN_TIME = 1e-10
MAX_TIME = 1e-7
TIME_P_FACTOR = 0.2
